import dataclasses
from typing import Any, Generic, Iterable, Optional, Sequence, Type, TypeVar, Union

from django.contrib.auth.models import AbstractUser, AnonymousUser
from graphql.type.definition import GraphQLResolveInfo
from strawberry.django.context import StrawberryDjangoContext
from strawberry.schema_directive import StrawberrySchemaDirective
from strawberry.types.info import Info
from typing_extensions import TypeAlias

_T = TypeVar("_T")

TypeOrSequence: TypeAlias = Union[_T, Sequence[_T]]
TypeOrIterable: TypeAlias = Union[_T, Iterable[_T]]
UserType: TypeAlias = Union[AbstractUser, AnonymousUser]
ResolverInfo: TypeAlias = Union[Info[StrawberryDjangoContext, Any], GraphQLResolveInfo]


# FIXME: Use schema_directive once it is properly typed
class SchemaDirective(Generic[_T], StrawberrySchemaDirective):
    wrap: Type[_T]
    instance: Optional[_T] = dataclasses.field(init=False)



from typing import Awaitable, Callable, Optional, Type, TypeVar, Union, cast, overload

from graphql.pyutils.is_awaitable import is_awaitable as _is_awaitable
from graphql.type.definition import GraphQLResolveInfo
from strawberry.types.info import Info
from strawberry.utils.await_maybe import AwaitableOrValue
from typing_extensions import TypeGuard

__all__ = [
    "is_awaitable",
    "resolve",
    "resolve_async",
]

_T = TypeVar("_T")
_R = TypeVar("_R")


def is_awaitable(
    value: AwaitableOrValue[_T],
    *,
    info: Optional[Union[Info, GraphQLResolveInfo]] = None,
) -> TypeGuard[Awaitable[_T]]:
    """Check if the given value is awaitable.

    Args:
        value:
            The value to check for awaitability
        info:
            Optional gql execution info. If present, will use its implementation
            of `is_awaitable`, which might have some optimizations. Otherwise
            will fallback to `inspect.is_awaitable`

    Returns:
        `True` if the value is awaitable, `False` otherwise

    """
    if info is not None:
        if isinstance(info, Info):
            info = info._raw_info
        return info.is_awaitable(value)
    return _is_awaitable(value)


async def resolve_async(
    value: Awaitable[_T],
    resolver: Callable[[_T], AwaitableOrValue[_R]],
    *,
    ensure_type: Optional[Type[_R]] = None,
    info: Optional[Union[Info, GraphQLResolveInfo]] = None,
) -> _R:
    """Call resolver with the awaited value's response.

    Args:
        value:
            The value to be awaited
        resolver:
            The resolver to be called after the value was awaited.
        ensure_type:
            Optional type to ensure that the retval is an instance of it.

    Returns:
        An `Awaitable` with the return value of `resolver(await value)`

    Raises:
        TypeError: If ensure_type was passed and the return value
        is not an instance of it (checked using `instance(retval, ensyure_type)`).

    """
    ret = resolver(await value)
    while is_awaitable(ret, info=info):
        ret = await ret
    if ensure_type is not None and not isinstance(ret, ensure_type):
        raise TypeError(f"{ensure_type} expected, found {repr(ret)}")

    # FIXME: Remove cast once pyright resolves the negative TypeGuard form
    ret = cast(_R, ret)

    return ret


@overload
def resolve(
    value: _T,
    resolver: Callable[[_T], _R],
    *,
    ensure_type: Optional[Type[_R]] = None,
) -> _R:
    ...


@overload
def resolve(
    value: AwaitableOrValue[_T],
    resolver: Callable[[_T], Awaitable[_R]],
    *,
    ensure_type: Optional[Type[_R]] = None,
    info: Optional[Union[Info, GraphQLResolveInfo]] = None,
) -> AwaitableOrValue[_R]:
    ...


@overload
def resolve(
    value: AwaitableOrValue[_T],
    resolver: Callable[[_T], _R],
    *,
    ensure_type: Optional[Type[_R]] = None,
    info: Optional[Union[Info, GraphQLResolveInfo]] = None,
) -> AwaitableOrValue[_R]:
    ...


def resolve(value, resolver, *, ensure_type=None, info=None):
    """Call resolver with the value's response.

    Args:
        value:
            The value to be passed to the resolver.
        info:
            The resolver to be called after the value was awaited.
        ensure_type:
            Optional type to ensure that the retval is an instance of it.
        info:
            Optional gql execution info. If present, will use its implementation
            of `is_awaitable`, which might have some optimizations. Otherwise
            will fallback to `inspect.is_awaitable`

    Returns:
        If the value is not awaitable, it will simply return `resolver(value)`.
        Otherwise, an `Awaitable` will be returned that, when awaited will return
        the return value of `resolver(await value)`.

    Raises:
        TypeError: If ensure_type was passed and the return value
        is not an instance of it (checked using `instance(retval, ensyure_type)`).

    """
    if is_awaitable(value, info=info):
        return resolve_async(value, resolver, info=info, ensure_type=ensure_type)

    ret = resolver(value)
    if ensure_type is not None and not isinstance(ret, ensure_type):
        raise TypeError(f"{ensure_type} expected, found {repr(ret)}")

    return ret



import dataclasses
import sys
from typing import Any, Dict, Iterable, Mapping
import weakref

from graphql.language.directive_locations import DirectiveLocation
from graphql.language.printer import print_ast
from graphql.type.definition import GraphQLArgument
from graphql.type.directives import GraphQLDirective
from graphql.utilities.ast_from_value import ast_from_value
from graphql.utilities.print_schema import print_directive
from strawberry import Schema, printer
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import UNSET, StrawberryArgument, is_unset
from strawberry.field import StrawberryField
from strawberry.private import is_private
from strawberry.schema.base import BaseSchema
from strawberry.schema_directive import StrawberrySchemaDirective

_original_print_schema = printer.print_schema
_directives = weakref.WeakKeyDictionary()


def _normalize_dataclasses(value: Any) -> Any:
    if dataclasses.is_dataclass(value):
        return dataclasses.asdict(value)
    if isinstance(value, Iterable):
        return [_normalize_dataclasses(v) for v in value]
    if isinstance(value, Mapping):
        return {k: _normalize_dataclasses(v) for k, v in value.items()}

    return value


def _print_schema_directive_arg(
    directive: StrawberrySchemaDirective,
    name: str,
    arg: GraphQLArgument,
):
    value = getattr(directive.instance, name, UNSET)
    if is_unset(value):
        return ""

    ast = ast_from_value(_normalize_dataclasses(value), arg.type)
    return ast and f"{name}: {print_ast(ast)}"


def _print_schema_directive_args(
    directive: StrawberrySchemaDirective,
    args: Dict[str, GraphQLArgument],
):
    printed = []
    for name, arg in args.items():
        p = _print_schema_directive_arg(directive, name, arg)
        if p:
            printed.append(p)

    if not printed:
        return ""

    return f'({", ".join(printed)})'


def _print_schema_directive(directive: StrawberrySchemaDirective, schema: Schema) -> str:
    cls = directive.wrap
    name_converter = schema.config.name_converter
    schema_converter = schema.schema_converter

    module = sys.modules[cls.__module__]

    _args: Dict[str, StrawberryArgument] = {}
    for field in dataclasses.fields(cls):
        if is_private(field.type):
            continue

        if isinstance(field, StrawberryField):
            default = field.default
        else:
            default = getattr(cls, field.name, UNSET)

        if default == dataclasses.MISSING:
            default = UNSET

        arg = StrawberryArgument(
            python_name=field.name,
            graphql_name=None,
            type_annotation=StrawberryAnnotation(
                annotation=field.type,
                namespace=module.__dict__,
            ),
            default=default,
        )
        _args[name_converter.from_argument(arg)] = arg

    args = {k: schema_converter.from_argument(v) for k, v in _args.items()}
    d = GraphQLDirective(
        name=name_converter.from_directive(directive),
        locations=[DirectiveLocation(loc.value) for loc in directive.locations],
        is_repeatable=False,
        args=args,
        description=directive.description,
    )
    _directives.setdefault(schema, set()).add(print_directive(d))

    return f" @{d.name}{_print_schema_directive_args(directive, args)}"


def print_schema(schema: BaseSchema) -> str:
    parts = [_original_print_schema(schema)]
    directives = _directives.get(schema)
    if directives:
        parts.insert(0, "\n\n".join(directives))
    return "\n\n".join(parts)


printer.print_schema_directive = _print_schema_directive
printer.print_schema = print_schema



from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from strawberry.arguments import UNSET
from strawberry.permission import BasePermission
from strawberry.schema_directive import StrawberrySchemaDirective
from strawberry.types.fields.resolver import StrawberryResolver

from strawberry_django_plus.utils.resolvers import async_unsafe

from . import relay

_T = TypeVar("_T")


class DjangoInputMutationField(relay.InputMutationField):
    """Relay Mutation field for django models.

    This is basically the same as `relay.InputMutationField`, but it ensure that
    the mutation resolver gets called in an async safe environment.

    Do not instantiate this directly. Instead, use `@gql.django.inpt_mutation`

    """

    def __call__(self, resolver: Callable[..., Iterable[relay.Node]]):
        return super().__call__(async_unsafe(resolver))  # type:ignore


@overload
def input_mutation(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
) -> _T:
    ...


@overload
def input_mutation(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
) -> Any:
    ...


@overload
def input_mutation(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
) -> DjangoInputMutationField:
    ...


def input_mutation(
    resolver=None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init=None,
) -> Any:
    """Annotate a property or a method to create an input mutation field.

    This is basically the same as `@relay.input_mutation`, but it ensure
    that the mutation resolver gets called in an async safe environment.

    """
    f = DjangoInputMutationField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        directives=directives or (),
    )
    if resolver is not None:
        f = f(resolver)
    return f



from typing import Dict, Mapping

from typing_extensions import TypeAlias

DictTree: TypeAlias = Dict[str, "DictTree"]


def dicttree_merge(dict1: DictTree, dict2: DictTree) -> DictTree:
    new = {
        **dict1,
        **dict2,
    }

    for k, v1 in dict1.items():
        if not isinstance(v1, Mapping):
            continue

        v2 = dict2.get(k)
        if isinstance(v2, Mapping):
            new[k] = dicttree_merge(v1, v2)

    for k, v2 in dict2.items():
        if k in new or not isinstance(v2, Mapping):
            continue

        v1 = dict1.get(k)
        if isinstance(v1, Mapping):
            new[k] = dicttree_merge(v1, v2)

    return new


def dicttree_intersect_diff(dict1: DictTree, dict2: DictTree) -> bool:
    for k in set(dict1) & set(dict2):
        v1 = dict1[k]
        v2 = dict2[k]

        if isinstance(v1, Mapping) and isinstance(v2, Mapping):
            v_intersect = dicttree_intersect_diff(v1, v2)
            if v_intersect:
                return True

        return v1 != v2

    return False



from strawberry import (
    ID,
    BasePermission,
    LazyType,
    Private,
    Schema,
    argument,
    auto,
    directive,
    enum,
    experimental,
    federation,
    field,
    input,
    interface,
    mutation,
    scalar,
    schema_directive,
    subscription,
    type,
    union,
)

from strawberry_django_plus import relay
from strawberry_django_plus.relay import (
    Connection,
    Node,
    connection,
    input_mutation,
    node,
)

from . import django

__all__ = [
    # strawberry
    "BasePermission",
    "ID",
    "LazyType",
    "Private",
    "Schema",
    "argument",
    "auto",
    "directive",
    "enum",
    "experimental",
    "federation",
    "field",
    "input",
    "interface",
    "mutation",
    "scalar",
    "schema_directive",
    "subscription",
    "type",
    "union",
    # strawberry_django_plus
    "relay",
    "Node",
    "Connection",
    "node",
    "connection",
    "input_mutation",
    "django",
]



import inspect

from strawberry import auto as _auto
from strawberry import object_type as _object_type
from strawberry.enum import EnumDefinition as _EnumDefinition
from strawberry.field import StrawberryField as _StrawberryField
from strawberry.schema_directive import (
    StrawberrySchemaDirective as _StrawberrySchemaDirective,
)
from strawberry.types.fields.resolver import StrawberryResolver as _StrawberryResolver
from strawberry_django import types as _types
from strawberry_django.fields import types as _ftypes

# Just import this for the monkey patch
from .utils.printer import print_schema  # noqa:F401

# Monkey patch strawberry_django to use strawberry's auto
_ftypes.auto = _auto
_types.auto = _auto  # type:ignore

_cls_docs = {}
_original_process_type = _object_type._process_type
_original_process_type = _object_type._process_type
_original_wrap_dataclass = _object_type._wrap_dataclass
_original_field_init = _StrawberryField.__init__
_original_field_call = _StrawberryField.__call__
_original_enum_init = _EnumDefinition.__init__
_original_schema_directive_init = _StrawberrySchemaDirective.__init__


def _get_doc(obj):
    if not obj.__doc__:
        return None
    return inspect.cleandoc(obj.__doc__)


def _process_type(cls, *args, **kwargs):
    if kwargs.get("description") is None:
        kwargs["description"] = _cls_docs.get(cls)
    ret = _original_process_type(cls, *args, **kwargs)
    for d in ret._type_definition.directives:
        d.instance.register(ret._type_definition)
    return ret


def _wrap_dataclass(cls):
    _cls_docs[cls] = _get_doc(cls)
    return _original_wrap_dataclass(cls)


def _field_init(self, *args, **kwargs):
    if kwargs.get("description") is None:
        base_resolver = kwargs.get("base_resolver")
        if base_resolver is not None:
            while isinstance(base_resolver, _StrawberryResolver):
                base_resolver = base_resolver.wrapped_func
            kwargs["description"] = _get_doc(base_resolver)
    ret = _original_field_init(self, *args, **kwargs)
    for d in self.directives:
        d.instance.register(self)
    return ret


def _field_call(self, resolver):
    ret = _original_field_call(self, resolver)
    if self.description is None:
        resolver = self.base_resolver
        while isinstance(resolver, _StrawberryResolver):
            resolver = resolver.wrapped_func
        self.description = _get_doc(resolver)
    return ret


def _enum_init(*args, **kwargs):
    if kwargs.get("description") is None:
        cls = kwargs.get("wrapped_cls")
        kwargs["description"] = _get_doc(cls)
    return _original_enum_init(*args, **kwargs)


def _schema_directive_init(self, *args, **kwargs):
    if kwargs.get("description") is None:
        cls = kwargs.get("wrap")
        kwargs["description"] = _get_doc(cls)
    return _original_schema_directive_init(self, *args, **kwargs)


_object_type._process_type = _process_type
_object_type._wrap_dataclass = _wrap_dataclass
_StrawberryField.__init__ = _field_init
_StrawberryField.__call__ = _field_call
_EnumDefinition.__init__ = _enum_init
_StrawberrySchemaDirective.__init__ = _schema_directive_init



import inspect
from typing import TYPE_CHECKING, Union

from django.db.models.fields import Field
from django.db.models.fields.reverse_related import ForeignObjectRel
from strawberry import enum
from strawberry_django.fields.types import (
    resolve_model_field_type as _resolve_model_field,
)

if TYPE_CHECKING:
    from strawberry_django_plus.type import StrawberryDjangoType

try:
    from django_choices_field import IntegerChoicesField, TextChoicesField
except ImportError:
    has_choices_field = False
else:
    has_choices_field = True


def resolve_model_field_type(
    model_field: Union[Field, ForeignObjectRel],
    django_type: "StrawberryDjangoType",
):
    if has_choices_field and isinstance(model_field, (TextChoicesField, IntegerChoicesField)):
        field_type = model_field.choices_enum
        enum_def = getattr(field_type, "_enum_definition", None)
        if enum_def is None:
            doc = field_type.__doc__ and inspect.cleandoc(field_type.__doc__)
            enum_def = enum(field_type, description=doc)._enum_definition
        return enum_def

    return _resolve_model_field(model_field, django_type)



import dataclasses
import types
from typing import Callable, Generic, Optional, Sequence, Type, TypeVar, Union

from django.db.models.base import Model
import strawberry
from strawberry import auto
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import UNSET
from strawberry.field import StrawberryField
from strawberry.schema_directive import StrawberrySchemaDirective
from strawberry.types.info import Info
from strawberry.unset import _Unset
from strawberry.utils.typing import __dataclass_transform__
from strawberry_django.fields.field import field as _field
from strawberry_django.filters import StrawberryDjangoFieldFilters
from strawberry_django.ordering import StrawberryDjangoFieldOrdering
from strawberry_django.pagination import StrawberryDjangoPagination
from strawberry_django.type import StrawberryDjangoType as _StraberryDjangoType
from strawberry_django.utils import get_annotations

from .field import StrawberryDjangoField, field
from .permissions import HasPermDirective
from .relay import Node, connection, node
from .utils.resolvers import (
    resolve_connection,
    resolve_model_id,
    resolve_model_node,
    resolve_model_nodes,
)

__all = [
    "StrawberryDjangoType",
    "type",
    "interface",
    "input",
]

_T = TypeVar("_T")
_O = TypeVar("_O", bound=type)
_M = TypeVar("_M", bound=Model)


def _get_fields(django_type: "StrawberryDjangoType"):
    origin = django_type.origin
    fields = {}

    # collect all annotated fields
    for name, annotation in get_annotations(origin).items():
        fields[name] = StrawberryDjangoField.from_django_type(
            django_type,
            name,
            type_annotation=annotation,
        )

    # collect non-annotated strawberry fields
    for name in dir(origin):
        if name in fields:
            continue

        attr = getattr(origin, name)
        if not isinstance(attr, StrawberryField):
            continue

        fields[name] = StrawberryDjangoField.from_django_type(django_type, name)

    return fields


def _has_own_node_resolver(cls, name: str) -> bool:
    resolver = getattr(cls, name, None)
    if resolver is None:
        return False

    if id(resolver.__func__) == id(getattr(Node, name).__func__):
        return False

    return True


def _get_filters(info: Info = None):
    if not info:
        return []

    return HasPermDirective.for_origin(info._field)


def _process_type(
    cls: _O,
    model: Type[Model],
    *,
    filters: Optional[StrawberryDjangoFieldFilters] = UNSET,
    pagination: Optional[StrawberryDjangoPagination] = UNSET,
    order: Optional[StrawberryDjangoFieldOrdering] = UNSET,
    **kwargs,
) -> _O:
    original_annotations = cls.__dict__.get("__annotations__", {})

    django_type = StrawberryDjangoType(
        origin=cls,
        model=model,
        is_input=kwargs.get("is_input", False),
        is_partial=kwargs.pop("partial", False),
        is_filter=kwargs.pop("is_filter", False),
        filters=filters,
        order=order,
        pagination=pagination,
    )

    fields = list(_get_fields(django_type).values())
    cls.__annotations__ = {}

    # update annotations and fields
    for f in fields:
        annotation = f.type_annotation.annotation if f.type_annotation is not None else f.type
        if annotation is None:
            annotation = StrawberryAnnotation(auto)

        cls.__annotations__[f.name] = annotation
        setattr(cls, f.name, f)

    # Make sure model is also considered a "virtual subclass" of cls
    if not hasattr(cls, "is_type_of"):
        cls.is_type_of = lambda obj, info: isinstance(obj, (cls, model))  # type:ignore

    # Default querying methods for relay
    if issubclass(cls, Node):
        if not _has_own_node_resolver(cls, "resolve_node"):
            cls.resolve_node = types.MethodType(resolve_model_node, cls)

        if not _has_own_node_resolver(cls, "resolve_nodes"):
            cls.resolve_nodes = types.MethodType(
                lambda *args, **kwargs: resolve_model_nodes(
                    *args,
                    filters=_get_filters(kwargs.get("info")),
                    **kwargs,
                ),
                cls,
            )

        if not _has_own_node_resolver(cls, "resolve_connection_resolver"):
            cls.resolve_connection = types.MethodType(
                lambda *args, **kwargs: resolve_connection(
                    *args,
                    filters=_get_filters(kwargs.get("info")),
                    **kwargs,
                ),
                cls,
            )

        if not _has_own_node_resolver(cls, "resolve_id"):
            cls.resolve_id = types.MethodType(
                lambda cls, root, *args, **kwargs: resolve_model_id(root),
                cls,
            )

    strawberry.type(cls, **kwargs)

    # restore original annotations for further use
    cls.__annotations__ = original_annotations
    cls._django_type = django_type  # type:ignore

    return cls


@dataclasses.dataclass
class StrawberryDjangoType(Generic[_O, _M], _StraberryDjangoType):
    """Strawberry django type metadata."""

    origin: _O
    model: Type[_M]
    is_input: bool
    is_filter: bool
    is_partial: bool
    order: Optional[Union[StrawberryDjangoFieldOrdering, _Unset]]
    filters: Optional[Union[StrawberryDjangoFieldFilters, _Unset]]
    pagination: Optional[Union[StrawberryDjangoPagination, _Unset]]


@__dataclass_transform__(
    order_default=True,
    field_descriptors=(StrawberryField, field, _field, node, connection),
)
def type(  # noqa:A001
    model: Type[Model],
    *,
    name: str = None,
    is_input: bool = False,
    is_interface: bool = False,
    description: str = None,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    extend: bool = False,
    filters: Optional[StrawberryDjangoFieldFilters] = UNSET,
    pagination: Optional[StrawberryDjangoPagination] = UNSET,
    order: Optional[StrawberryDjangoFieldOrdering] = UNSET,
) -> Callable[[_T], _T]:
    """Annotates a class as a Django GraphQL type.

    Examples:
        It can be used like this:

        >>> @gql.django.type(SomeModel)
        ... class X:
        ...     some_field: gql.auto
        ...     otherfield: str = gql.django.field()

    """

    def wrapper(cls):
        return _process_type(
            cls,
            model,
            name=name,
            is_input=is_input,
            is_interface=is_interface,
            description=description,
            directives=directives,
            extend=extend,
            filters=filters,
            pagination=pagination,
            order=order,
        )

    return wrapper


@__dataclass_transform__(
    order_default=True,
    field_descriptors=(StrawberryField, field, _field, node, connection),
)
def interface(
    model: Type[Model],
    *,
    name: str = None,
    description: str = None,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
) -> Callable[[_T], _T]:
    """Annotates a class as a Django GraphQL interface.

    Examples:
        It can be used like this:

        >>> @gql.django.interface(SomeModel)
        ... class X:
        ...     some_field: gql.auto
        ...     otherfield: str = gql.django.field()

    """
    return type(
        model,
        is_interface=True,
        name=name,
        description=description,
        directives=directives,
    )


@__dataclass_transform__(
    order_default=True,
    field_descriptors=(StrawberryField, field, _field, node, connection),
)
def input(  # noqa:A001
    model: Type[Model],
    *,
    name: str = None,
    description: str = None,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    partial: bool = False,
) -> Callable[[_T], _T]:
    """Annotates a class as a Django GraphQL interface.

    Examples:
        It can be used like this:

        >>> @gql.django.input(SomeModel)
        ... class X:
        ...     some_field: gql.auto
        ...     otherfield: str = gql.django.field()

    """
    return type(
        model,
        is_input=True,
        name=name,
        description=description,
        directives=directives,
    )



import contextlib
import dataclasses
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from django.db import models
from django.db.models import Prefetch
from django.db.models.fields.reverse_related import (
    ManyToManyRel,
    ManyToOneRel,
    OneToOneRel,
)
from django.db.models.manager import BaseManager
from django.db.models.query import QuerySet
from graphql.type.definition import GraphQLResolveInfo, get_named_type
from strawberry.extensions.base_extension import Extension
from strawberry.schema.schema import Schema
from strawberry.types.execution import ExecutionContext
from strawberry.types.info import Info
from strawberry.types.nodes import InlineFragment, Selection, convert_selections
from strawberry.types.types import TypeDefinition
from strawberry.utils.await_maybe import AwaitableOrValue
from strawberry_django.fields.types import resolve_model_field_name
from typing_extensions import TypeAlias

from .descriptors import ModelProperty
from .relay import Connection, Edge, NodeType
from .utils import resolvers
from .utils.inspect import (
    PrefetchInspector,
    get_model_fields,
    get_possible_type_definitions,
    get_selections,
)
from .utils.typing import TypeOrSequence

__all__ = [
    "OptimizerConfig",
    "DjangoOptimizerExtension",
    "OptimizerStore",
    "PrefetchType",
    "optimize",
]

_sentinel = object()
_T = TypeVar("_T")
_M = TypeVar("_M", bound=models.Model)

PrefetchType: TypeAlias = Union[str, Prefetch]


def _get_model_hints(
    model: Type[models.Model],
    schema: Schema,
    type_def: TypeDefinition,
    selection: Selection,
    *,
    config: Optional["OptimizerConfig"] = None,
    prefix: str = "",
    model_cache: Optional[Dict[Type[models.Model], List[Tuple[int, "OptimizerStore"]]]] = None,
    level: int = 0,
) -> "OptimizerStore":
    store = OptimizerStore()
    model_cache = model_cache or {}
    typename = schema.config.name_converter.from_object(type_def)

    # In case this is a relay field, find the selected edges/nodes, the selected fields
    # are actually inside edges -> node selection...
    if type_def.concrete_of and issubclass(type_def.concrete_of.origin, Connection):
        n_type = type_def.type_var_map[NodeType]
        n_type_def = cast(TypeDefinition, n_type._type_definition)  # type:ignore

        for edges in get_selections(selection, typename=typename).values():
            if edges.name != "edges":
                continue

            e_type = Edge._type_definition.resolve_generic(Edge[n_type])  # type:ignore
            e_typename = schema.config.name_converter.from_object(e_type._type_definition)
            for node in get_selections(edges, typename=e_typename).values():
                if node.name != "node":
                    continue

                store |= _get_model_hints(
                    model=model,
                    schema=schema,
                    type_def=n_type_def,
                    selection=node,
                    config=config,
                    prefix=prefix,
                    model_cache=model_cache,
                    level=level,
                )

        return store

    fields = {schema.config.name_converter.get_graphql_name(f): f for f in type_def.fields}
    model_fields = get_model_fields(model)

    # Make sure that the model's pk is always selected when using only
    pk = model._meta.pk
    if pk is not None:
        store.only.append(pk.attname)

    for f_selection in get_selections(selection, typename=typename).values():
        field = fields[f_selection.name]

        # Add annotations from the field if they exist
        field_store = getattr(field, "store", None)
        if field_store is not None:
            store |= field_store.with_prefix(prefix) if prefix else field_store

        # Then from the model property if one is defined
        model_attr = getattr(model, field.python_name, None)
        if model_attr is not None and isinstance(model_attr, ModelProperty):
            attr_store = model_attr.store
            store |= attr_store.with_prefix(prefix) if prefix else attr_store

        # Lastly, from the django field itself
        model_fieldname: str = getattr(field, "django_name", field.python_name)
        model_field = model_fields.get(model_fieldname, None)
        if model_field is not None:
            path = f"{prefix}{model_fieldname}"

            if isinstance(model_field, (models.ForeignKey, OneToOneRel)):
                store.only.append(path)
                store.select_related.append(path)

                # If adding a reverse relation, make sure to select its pointer to us,
                # or else this might causa a refetch from the database
                if isinstance(model_field, OneToOneRel):
                    remote_field = model_field.remote_field
                    store.only.append(f"{path}__{resolve_model_field_name(remote_field)}")

                for f_type_def in get_possible_type_definitions(field.type):
                    f_model = model_field.related_model
                    f_store = _get_model_hints(
                        f_model,
                        schema,
                        f_type_def,
                        f_selection,
                        config=config,
                        model_cache=model_cache,
                        level=level + 1,
                    )
                    model_cache.setdefault(f_model, []).append((level, f_store))
                    store |= f_store.with_prefix(f"{path}__")
            elif isinstance(model_field, (models.ManyToManyField, ManyToManyRel, ManyToOneRel)):
                f_types = list(get_possible_type_definitions(field.type))
                if len(f_types) > 1:
                    # This might be a generic foreign key. In this case, just prefetch it
                    store.prefetch_related.append(model_fieldname)
                elif len(f_types) == 1:
                    remote_field = model_field.remote_field
                    f_store = _get_model_hints(
                        remote_field.model,
                        schema,
                        f_types[0],
                        f_selection,
                        config=config,
                        model_cache=model_cache,
                        level=level + 1,
                    )
                    if (config is None or config.enable_only) and f_store.only:
                        # If adding a reverse relation, make sure to select its pointer to us,
                        # or else this might causa a refetch from the database
                        f_store.only.append(cast(str, resolve_model_field_name(remote_field)))

                    model_cache.setdefault(remote_field.model, []).append((level, f_store))

                    f_qs = f_store.apply(remote_field.model.objects.all(), config=config)
                    f_prefetch = Prefetch(path, queryset=f_qs)
                    f_prefetch._optimizer_sentinel = _sentinel  # type:ignore
                    store.prefetch_related.append(f_prefetch)
            else:
                store.only.append(path)

    # DJango keeps track of known fields. That means that if one model select_related or
    # prefetch_related another one, and later another one select_related or prefetch_related the
    # model again, if the used fields there where not optimized in this call django would have
    # to fetch those again. By mergint those with us we are making sure to avoid that
    for inner_level, inner_store in model_cache.get(model, []):
        if inner_level > level and inner_store:
            # We only want the only/select_related from this. prefetch_related is something else
            store.only.extend(inner_store.only)
            store.select_related.extend(inner_store.select_related)

    return store


def optimize(
    qs: Union[QuerySet[_M], BaseManager[_M]],
    info: Union[GraphQLResolveInfo, Info],
    *,
    config: Optional["OptimizerConfig"] = None,
    store: Optional["OptimizerStore"] = None,
) -> QuerySet[_M]:
    """Optimize the given queryset considering the gql info.

    This will look through the gql selections, fields and model hints and apply
    `only`, `select_related` and `prefetch_related` optimizations according those
    on the `QuerySet`_.

    Note:
        This do not execute the queryset, it only optimizes it for when it is actually
        executed.

        It will also avoid doing any extra optimization if the queryset already has
        cached results in it, to avoid triggering extra queries later.

    Args:
        qs:
            The queryset to be optimized
        info:
            The current field execution info
        config:
            Optional config to use when doing the optimization
        config:
            Optional initial store to use for the optimization

    Returns:
        The optimized queryset

    .. _QuerySet:
        https://docs.djangoproject.com/en/dev/ref/models/querysets/

    """
    if isinstance(qs, BaseManager):
        qs = cast(QuerySet[_M], qs.all())

    # If the queryset already has cached results, just return it
    if qs._result_cache:  # type:ignore
        return qs

    if isinstance(info, Info):
        info = info._raw_info
    config = config or OptimizerConfig()
    store = store or OptimizerStore()
    schema = cast(Schema, info.schema._strawberry_schema)  # type:ignore

    field_name = info.field_name
    gql_type = get_named_type(info.return_type)
    strawberry_type = schema.get_type_by_name(gql_type.name)
    if strawberry_type is None:
        return qs

    for type_def in get_possible_type_definitions(strawberry_type):
        selection = next(
            (
                s
                for s in convert_selections(info, info.field_nodes)
                if not isinstance(s, InlineFragment) and s.name == field_name
            ),
            None,
        )
        if not selection:
            continue

        store |= _get_model_hints(
            qs.model,
            schema,
            type_def,
            selection,
            config=config,
        )

    # Nothing found do optimize, just skip this...
    if not store:
        return qs

    return store.apply(qs, config=config)


@dataclasses.dataclass
class OptimizerConfig:
    """Django optimization configuration.

    Attributes:
        enable_only:
            Enable `QuerySet.only` optimizations
        enable_select_related:
            Enable `QuerySet.select_related` optimizations
        enable_prefetch_related:
            Enable `QuerySet.prefetch_related` optimizations

    """

    enable_only: bool = dataclasses.field(default=True)
    enable_select_related: bool = dataclasses.field(default=True)
    enable_prefetch_related: bool = dataclasses.field(default=True)


@dataclasses.dataclass
class OptimizerStore:
    """Django optimization store.

    Attributes:
        only:
            Set of values to optimize using `QuerySet.only`
        selected:
            Set of values to optimize using `QuerySet.select_related`
        prefetch_related:
            Set of values to optimize using `QuerySet.prefetch_related`

    """

    only: List[str] = dataclasses.field(default_factory=list)
    select_related: List[str] = dataclasses.field(default_factory=list)
    prefetch_related: List[PrefetchType] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return any([self.only, self.select_related, self.prefetch_related])

    def __ior__(self, other: "OptimizerStore"):
        self.only.extend(other.only)
        self.select_related.extend(other.select_related)
        self.prefetch_related.extend(other.prefetch_related)
        return self

    def __or__(self, other: "OptimizerStore"):
        return self.copy().__ior__(other)

    def copy(self):
        return self.__class__(
            only=self.only[:],
            select_related=self.select_related[:],
            prefetch_related=self.prefetch_related[:],
        )

    @classmethod
    def with_hints(
        cls,
        *,
        only: Optional[TypeOrSequence[str]] = None,
        select_related: Optional[TypeOrSequence[str]] = None,
        prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    ):
        return cls(
            only=[only] if isinstance(only, str) else list(only or []),
            select_related=(
                [select_related] if isinstance(select_related, str) else list(select_related or [])
            ),
            prefetch_related=(
                [prefetch_related]
                if isinstance(prefetch_related, (str, Prefetch))
                else list(prefetch_related or [])
            ),
        )

    def with_prefix(self, prefix: str):
        prefetch_related = []
        for p in self.prefetch_related:
            if isinstance(p, str):
                prefetch_related.append(f"{prefix}{p}")
            elif isinstance(p, Prefetch):
                p.add_prefix(prefix)
                prefetch_related.append(p)
            else:  # pragma:nocover
                raise AssertionError(f"Unexpected prefetch type {repr(p)}")

        return self.__class__(
            only=[f"{prefix}{i}" for i in self.only],
            select_related=[f"{prefix}{i}" for i in self.select_related],
            prefetch_related=prefetch_related,
        )

    def apply(
        self,
        qs: QuerySet[_M],
        *,
        config: Optional[OptimizerConfig] = None,
    ) -> QuerySet[_M]:
        config = config or OptimizerConfig()

        if config.enable_prefetch_related and self.prefetch_related:
            # Add all str at the same time to make it easier to handle Prefetch below
            to_prefetch: Dict[str, PrefetchType] = {
                p: p for p in self.prefetch_related if isinstance(p, str)
            }

            for p in self.prefetch_related:
                # Already added above
                if isinstance(p, str):
                    continue

                path = cast(str, p.prefetch_to)  # type:ignore
                existing = to_prefetch.get(path)
                # The simplest case. The prefetch doesn't exist or is a string.
                # In this case, just replace it.
                if not existing or isinstance(existing, str):
                    to_prefetch[path] = p
                    continue

                p1 = PrefetchInspector(existing)
                p2 = PrefetchInspector(p)
                if getattr(existing, "_sentinel", None) is _sentinel:
                    ret = p1.merge(p2, allow_unsafe_ops=True)
                elif getattr(p, "_sentinel", None) is _sentinel:
                    ret = p2.merge(p1, allow_unsafe_ops=True)
                else:
                    # The order here doesn't matter
                    ret = p1.merge(p2)

                to_prefetch[path] = ret.prefetch

            qs = qs.prefetch_related(*to_prefetch.values())

        if config.enable_select_related and self.select_related:
            qs = qs.select_related(*self.select_related)

        if config.enable_only and self.only:
            qs = qs.only(*self.only)

        return qs


class DjangoOptimizerExtension(Extension):
    """Automatically optimize returned querysets from internal resolvers.

    Attributes:
        enable_only_optimization:
            Enable `QuerySet.only` optimizations
        enable_select_related_optimization:
            Enable `QuerySet.select_related` optimizations
        enable_prefetch_related_optimization:
            Enable `QuerySet.prefetch_related` optimizations

    Examples:
        Add the following to your schema configuration.

        >>> import strawberry
        >>> from strawberry_django_plus.optimizer import DjangoOptimizerExtension
        ...
        >>> schema = strawberry.Schema(
        ...     Query,
        ...     extensions=[
        ...         DjangoOptimizerExtension(),
        ...     ]
        ... )

    """

    _disabled: ClassVar[bool] = False

    def __init__(
        self,
        *,
        enable_only_optimization: bool = True,
        enable_select_related_optimization: bool = True,
        enable_prefetch_related_optimization: bool = True,
        execution_context: ExecutionContext = None,
    ):
        super().__init__(execution_context=execution_context)  # type:ignore
        self._config = OptimizerConfig(
            enable_only=enable_only_optimization,
            enable_select_related=enable_select_related_optimization,
            enable_prefetch_related=enable_prefetch_related_optimization,
        )

    @classmethod
    @contextlib.contextmanager
    def disable(cls):
        cls._disabled = True
        yield
        cls._disabled = False

    def on_request_start(self) -> AwaitableOrValue[None]:
        if self._disabled:
            return

        self.execution_context.context._django_optimizer_config = self._config

    def resolve(
        self,
        _next: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        *args,
        **kwargs,
    ) -> AwaitableOrValue[Any]:
        ret = _next(root, info, *args, **kwargs)

        if self._disabled:
            return ret

        if isinstance(ret, (BaseManager, QuerySet)):
            if isinstance(ret, BaseManager):
                ret = ret.all()
            if not ret._result_cache:  # type:ignore
                return resolvers.resolve_qs(optimize(qs=ret, info=info, config=self._config))

        return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strawberry_django_plus',
 'strawberry_django_plus.gql',
 'strawberry_django_plus.integrations',
 'strawberry_django_plus.management.commands',
 'strawberry_django_plus.middlewares',
 'strawberry_django_plus.utils']

package_data = \
{'': ['*'], 'strawberry_django_plus': ['templates/strawberry_django_plus/*']}

install_requires = \
['django>=3.2',
 'strawberry-graphql-django>=0.2.5',
 'strawberry-graphql>=0.93.4',
 'typing-extensions>=4.0.1']

setup_kwargs = {
    'name': 'strawberry-django-plus',
    'version': '0.4.2',
    'description': 'Enhanced Strawberry GraphQL integration with Django',
    'long_description': '# strawberry-django-plus\n\n[![build status](https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2Fblb-ventures%2Fstrawberry-django-plus%2Fbadge%3Fref%3Dmaster&style=flat)](https://actions-badge.atrox.dev/blb-ventures/straw/goto?ref=master)\n[![coverage](https://img.shields.io/codecov/c/github/blb-ventures/strawberry-django-plus.svg)](https://codecov.io/gh/blb-ventures/strawberry-django-plus)\n[![PyPI version](https://img.shields.io/pypi/v/strawberry-django-plus.svg)](https://pypi.org/project/strawberry-django-plus/)\n![python version](https://img.shields.io/pypi/pyversions/strawberry-django-plus.svg)\n![django version](https://img.shields.io/pypi/djversions/strawberry-django-plus.svg)\n\nEnhanced Strawberry integration with Django.\n\nBuilt on top of [strawberry-django](https://github.com/strawberry-graphql/strawberry-graphql-django)\nintegration, enhancing its overall functionality.\n\n## Features\n\n- All of supported features by `strawberry-django`.\n- Extension that [automatically optimize queries](#automatic-queryset-optimization) (using\n  `only`/`select_related`/`prefetch_related`) to solve graphql `N+1` problems, with support for\n  fragment spread, inline fragments, `@include`/`@skip` directives, prefetch merging, etc\n- Improved sync/async resolver that priorizes the model\'s cache to avoid have to use\n  [sync_to_async](https://docs.djangoproject.com/en/4.0/topics/async/#asgiref.sync.sync_to_async)\n  when not needed.\n- Support for [Django choices enums using](#django-choices-enums) (requires\n  [django-choices-field](https://github.com/bellini666/django-choices-field))\n- [Permissioning](#permissioning) using schema directives with the default\n  [django authentication system](https://docs.djangoproject.com/en/4.0/topics/auth/default/),\n  with support for both direct and per-object permission checking for backends that\n  support those (e.g. [django-guardian](https://django-guardian.readthedocs.io/en/stable])).\n- [Relay support](#relay-support) for queries, connections and input mutations.\n- [Django Debug Toolbar integration](#django-debug-toolbar-integration) with graphiql to\n  display metrics like SQL queries\n- (Coming Soon...) Improved Django mutations with automatic validation errors integration.\n- A well typed and documented API.\n\n## Installation\n\nInstall it with pip:\n\n```shell\npip install strawberry-django-plus\n```\n\n## Introduction\n\nSince this lib has a long name, it does provide a shortcut called `gql` where all of\nstrawberry\'s API and ours can be accessed.\n\n```python\nfrom strawberry_django_plus import gql\n\n# All strawberry\'s base api can be found directly on gql, like:\ngql.type  # same as strawberry.type\ngql.field  # same as strawberry.field\n...\n\n# The strawberry-django API and our custom implementation can be found on gql.django, like:\ngql.django.type\ngql.django.field\n...\n\n# We also have a custom relay implementation in here:\ngql.relay\n```\n\n## How To\n\n### Automatic QuerySet Optimization\n\nThe automatic optimization is enabled by adding the `DjangoOptimizerExtension` to your\nstrawberry\'s schema config.\n\n```python\nimport strawberry\nfrom strawberry_django_plus.optimizer import DjangoOptimizerExtension\n\nschema = strawberry.Schema(\n    Query,\n    extensions=[\n        # other extensions...\n        DjangoOptimizerExtension,\n    ]\n)\n```\n\nNow consider the following:\n\n```python\n# models.py\n\nclass Artist(models.Model):\n    name = models.CharField()\n\nclass Album(models.Moodel):\n    name = models.CharField()\n    release_date = models.DateTimeField()\n    artist = models.ForeignKey("Artist", related_name="albuns")\n\nclass Song(models.Model):\n    name = model.CharField()\n    duration = models.DecimalField()\n    album = models.ForeignKey("Album", related_name="songs")\n\n# schema.py\nfrom strawberry_django_plus import gql\n\n@gql.django.type(Artist)\nclass ArtistType:\n    name: auto\n    albums: "List[AlbumType]"\n\n@gql.django.type(Album)\nclass AlbumType:\n    name: auto\n    release_date: auto\n    artist: ArtistType\n    songs: "List[SongType]"\n\n@gql.django.type(Song)\nclass SongType:\n    name: auto\n    duration: auto\n    album_type: AlbumType\n\n@gql.type\nclass Query:\n    artist: Artist = gql.django.field()\n    songs: List[SongType] = gql.django.field()\n```\n\nQuerying the artist field:\n\n```gql\n{\n  artist {\n    id\n    name\n    albums {\n      id\n      name\n      songs {\n        id\n        name\n      }\n    }\n  }\n}\n```\n\n```python\n# This will generate a query like:\nArtist.objects.all().only("id", "name").prefetch_related(\n    Prefetch(\n        "albums",\n        queryset=Album.objects.all().only("id", "name").prefetch_related(\n            "songs",\n            Song.objects.all().only("id", "name"),\n        )\n    ),\n)\n```\n\nAnd querying the songs list:\n\n```gql\n{\n  song {\n    id\n    album\n    id\n    name\n    artist {\n      id\n      name\n      albums {\n        id\n        name\n        release_date\n      }\n    }\n  }\n}\n```\n\n```python\n# This will generate a query like:\nSong.objects.all().only(\n    "id",\n    "album",\n    "album__id",\n    "album__name",\n    "album__release_date",  # Note about this below\n    "album__artist",\n    "album__artist__id",\n).select_related(\n    "album",\n    "album__artist",\n).prefetch_related(\n    "album__artist__albums",\n    Prefetch(\n        "albums",\n        Album.objects.all().only("id", "name", "release_date"),\n    )\n)\n```\n\nNote that even though `album__release_date` field was not selected here, it got selected\nin the prefetch query later. Since Django caches known objects, we have to select it here or\nelse it would trigger extra queries latter.\n\nIt is also possible to include hints for non-model fields using the field api or even our\n`@model_property` (or its cached variation, `@cached_model_property`) decorator on the model\nitself, for people who likes to keep all the business logic at the model.\n\nFor example, the following will automatically optimize `only` and `select_related` if that\nfield gets selected:\n\n```python\nfrom strawberry_django_plus import gql\n\nclass Song(models.Model):\n    name = models.CharField()\n\n    @gql.model_property(only=["name", "album__name"], select_related=["album"])\n    def name_with_album(self) -> List[str]:\n        return f"{self.album.name}: {self.name}"\n\n@gql.django.type(Song)\nclass SongType:\n    name: auto\n    name_with_album: str\n```\n\nAnother option would be to define that on the field itself:\n\n```python\n@gql.django.type(Song)\nclass SongType:\n    name: auto\n    name_with_album: str = gql.django.field(\n        only=["name", "album__name"],\n        select_related=["album"],\n    )\n```\n\n### Django Choices Enums\n\nConvert choices fields into GraphQL enums by using\n[Django Choices Field](https://github.com/bellini666/django-choices-field) extension.\n\n```python\nfrom django_choices_field import TexChoicesField\n\nclass Song(models.Model):\n    class Genre(models.TextChoices):\n        ROCK = "rock", "Rock\'n\'Roll"\n        METAL = "metal", "Metal"\n        OTHERS = "others", "Who Cares?"\n\n    genre = TextChoicesField(choices_enum=Genre)\n```\n\nIn that example, a new enum called `Genre` will be created and be used for queries\nand mutations.\n\nIf you want to name it differently, decorate the class with `@gql.enum` with your preferred\nname so that strawberry-django-plus will not try to register it again.\n\n### Permissioning\n\nPermissioning is done using schema directives by applying them to the fields that requires\npermission checking.\n\nFor example:\n\n```python\n@strawberry.type\nclass SomeType:\n    login_required_field: RetType = strawberry.field(\n        directives=[IsAuthenticated()],\n    )\n    perm_required_field: OtherType = strawberry.field(\n        directives=[HasPerm("some_app.some_perm")],\n    )\n    obj_perm_required_field: OtherType = strawberry.field(\n        directives=[HasObjPerm("some_app.some_perm")],\n    )\n```\n\n- `login_required_field` will check if the user is authenticated\n- `perm_required_field` will check if the user has `"some_app.some_perm"` permission\n- `obj_perm_required_field` will check the permission for the resolved value\n\nAvailable options are:\n\n- `IsAuthenticated`: Checks if the user is authenticated (`user.is_autenticated`)\n- `IsStaff`: Checks if the user is a staff member (`user.is_staff`)\n- `IsSuperuser`: Checks if the user is a superuser (`user.is_superuser`)\n- `HasPerm(perms: str, list[str], any: bool = True)`: Checks if the user has any or all of\n  the given permissions (`user.has_perm(perm)`)\n- `HasRootPerm(perms: str | list[str], any: bool = True)`: Checks if the user has any or all\n  of the given permissions for the root of that field (`user.has_perm(perm, root)`)\n- `HasObjPerm(perms: str | list[str], any: bool = True)`: Resolves the retval and then\n  checks if the user has any or all of the given permissions for that specific value\n  (`user.has_perm(perm, retval)`). Note that if the return value is a list, this directive\n  will filter the return value, removing objects that fails the check (check below for more\n  information regarding other possibilities).\n\nThere are some important notes regarding how the directives handle the return value:\n\n- If the user passes the check, the retval is returned normally\n- If the user fails the check:\n  - If the return type was `Optional`, it returns `None`\n  - If the return type was a `List`, it returns an empty list\n  - If the return type was a relay `Connection`, it returns an empty `Connection`\n  - Otherwise, it raises a `PermissionError` for that resolver\n\nNote that since `strawberry` doesn\'t support resolvers for schema directives, it is necessary\nto use this lib\'s custom extension that handles the resolution of those and any other custom\ndefined schema directive inherited from `strawberry_django_plus.directives.SchemaDirectiveResolver`:\n\n```python\nimport strawberry\nfrom strawberry_django_plus.directives import SchemaDirectiveExtension\n\nschema = strawberry.Schema(\n    Query,\n    extensions=[\n        SchemaDirectiveExtension,\n        # other extensions...\n    ]\n)\n```\n\n### Relay Support\n\nWe have a custom [relay spec](https://relay.dev/docs/guides/graphql-server-specification/)\nimplementation. It is not tied to Django at all to allow its usage with other types.\n\nIt provides types and fields for node and connection querying. For example:\n\n```python\n# schema.py\nfrom strawberry_django_plus import gql\nfrom strawberry_django_plus.gql import relay\n\n@gql.type\nclass Fruit(relay.Node):\n    name: str\n\n    def resolve_node(cls, node_id, info, required=False):\n        ...\n\n    def resolve_nodes(cls, node_id, info, node_ids=False):\n        ...\n\n\n@gql.type\nclass Query:\n    fruit: Optional[Fruit] = relay.node()\n    fruits_connection: relay.Connection[Fruit] = relay.connection()\n\n    @relay.connection\n    def fruits_connection_filtered(self, name_startswith: str) -> Iterable[Fruit]:\n        # Note that this resolver is special. It should not resolve the connection, but\n        # the iterable of nodes itself. Thus, any arguments defined here will be appended\n        # to the query, and the pagination of the iterable returned here will be\n        # automatically handled.\n        ...\n```\n\nThis will generate a schema like this:\n\n```gql\ninterface Node {\n  id: GlobalID!\n}\n\ntype Fruit implements Node {\n  id: GlobalID!\n  name: String!\n}\n\ntype FruitEdge implements Node {\n  cursor: String!\n  node: Fruit\n}\n\ntype FruitConnection {\n  edges: [ShipEdge!]!\n  pageInfo: PageInfo!\n}\n\ntype PageInfo {\n  hasNextPage: Boolean!\n  hasPreviousPage: Boolean!\n  startCursor: String\n  endCursor: String\n}\n\ntype Query {\n  fruit(id: GlobalID!): Fruit\n  fruits_connection(\n    before: String\n    after: String\n    first: Int\n    last: Int\n  ): FruitConnection\n  fruits_connection_filtered(\n    before: String\n    after: String\n    first: Int\n    last: Int\n    nameStartswith: String!\n  ): FruitConnection\n}\n```\n\nIt is expected that types implementing the `Node` interface define some methods, like\n`resolve_nodes` and `resolve_node`. Take a look at\n[the documentation](.strawberry_django_plus.relay.py) for more information.\n\nAlso note that Django fields created with `@gql.django.type` automatically implements\nall of the required methods when the type inherits from `Node`.\n\nThis module also exposes a mutation that converts all of its arguments to a single input.\nFor example:\n\n```python\n@gql.type\nclass Mutation:\n    @relay.input_mutation\n    def create_fruit(name: str) -> Fruit:\n        ....\n```\n\nWill generate those types:\n\n```gql\ninput CreateFruitInput {\n  name: String!\n}\n\ntype Mutation {\n  createFruit(input: CreateFruitInput!): Fruit\n}\n```\n\n### Django Debug Toolbar Integration\n\nInstall [Django Debug Toolbar](https://github.com/jazzband/django-debug-toolbar)\nand change its middleware from:\n\n```python\nMIDDLEWARE = [\n    ...\n    "debug_toolbar.middleware.DebugToolbarMiddleware",\n    ...\n]\n```\n\nTo:\n\n```python\nMIDDLEWARE = [\n    ...\n    "strawberry_django_plus.middlewares.debug_toolbar.DebugToolbarMiddleware",\n    ...\n]\n```\n\n## Contributing\n\nWe use [poetry](https://github.com/sdispater/poetry) to manage dependencies, to\nget started follow these steps:\n\n```shell\ngit clone https://github.com/blb-ventures/strawberry-django-plus\ncd strawberry\npoetry install\npoetry run pytest\n```\n\nThis will install all the dependencies (including dev ones) and run the tests.\n\n### Pre commit\n\nWe have a configuration for\n[pre-commit](https://github.com/pre-commit/pre-commit), to add the hook run the\nfollowing command:\n\n```shell\npre-commit install\n```\n\n## Licensing\n\nThe code in this project is licensed under MIT license. See [LICENSE](./LICENSE)\nfor more information.\n',
    'author': 'Thiago Bellini Ribeiro',
    'author_email': 'thiago@bellini.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/blb-ventures/strawberry-django-plus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import dataclasses
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import QuerySet
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.fields.reverse_related import ManyToManyRel, ManyToOneRel
from django.db.models.query_utils import DeferredAttribute
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import UNSET, is_unset
from strawberry.permission import BasePermission
from strawberry.schema_directive import StrawberrySchemaDirective
from strawberry.types.fields.resolver import StrawberryResolver
from strawberry.types.info import Info
from strawberry_django.fields.field import (
    StrawberryDjangoField as _StrawberryDjangoField,
)
from strawberry_django.fields.types import (
    get_model_field,
    is_auto,
    is_optional,
    resolve_model_field_name,
)
from strawberry_django.utils import is_similar_django_type
from typing_extensions import Self

from .descriptors import ModelProperty
from .optimizer import OptimizerStore, PrefetchType
from .permissions import HasPermDirective, is_perm_safe, perm_safe
from .types import resolve_model_field_type
from .utils import resolvers
from .utils.typing import TypeOrSequence

if TYPE_CHECKING:
    from .type import StrawberryDjangoType

__all__ = [
    "StrawberryDjangoField",
    "field",
]

_T = TypeVar("_T")
_M = TypeVar("_M", bound=models.Model)


class StrawberryDjangoField(_StrawberryDjangoField):
    """A strawberry field for django attributes.

    Do not instantiate this directly. Instead, use `@field` decorator.

    """

    store: OptimizerStore

    def __init__(self, *args, **kwargs):
        self.store = OptimizerStore.with_hints(
            only=kwargs.pop("only", None),
            select_related=kwargs.pop("select_related", None),
            prefetch_related=kwargs.pop("prefetch_related", None),
        )
        super().__init__(*args, **kwargs)

    @cached_property
    def model(self) -> Type[models.Model]:
        model = self.django_model
        if model:
            return model

        origin = self.origin_django_type or self.origin._django_type
        return origin.model

    @classmethod
    def from_django_type(
        cls,
        django_type: "StrawberryDjangoType",
        name: str,
        *,
        type_annotation: Optional[StrawberryAnnotation] = None,
    ) -> Self:
        origin = django_type.origin

        attr = getattr(origin, name, UNSET)
        if is_unset(attr):
            attr = getattr(cls, "__dataclass_fields__", {}).get(name, UNSET)
        if attr is dataclasses.MISSING:
            attr = UNSET

        if isinstance(attr, cls) and not attr.origin_django_type:
            field = cast(Self, attr)
        elif isinstance(attr, dataclasses.Field):
            default = getattr(attr, "default", UNSET)
            if default is dataclasses.MISSING:
                default = UNSET

            default_factory = getattr(attr, "default_factory", UNSET)
            if default_factory is dataclasses.MISSING:
                default_factory = UNSET

            if type_annotation is None:
                type_annotation = getattr(attr, "type_annotation", None)
            if type_annotation is None:
                type_annotation = StrawberryAnnotation(attr.type)

            field = cls(
                django_name=getattr(attr, "django_name", attr.name),
                graphql_name=getattr(attr, "graphql_name", None),
                origin=getattr(attr, "origin", None),
                is_subscription=getattr(attr, "is_subscription", False),
                description=getattr(attr, "description", None),
                base_resolver=getattr(attr, "base_resolver", None),
                permission_classes=getattr(attr, "permission_classes", ()),
                default=default,
                default_factory=default_factory,
                deprecation_reason=getattr(attr, "deprecation_reason", None),
                directives=getattr(attr, "directives", ()),
                type_annotation=type_annotation,
            )
        elif isinstance(attr, StrawberryResolver):
            field = cls(base_resolver=attr)
        elif callable(attr):
            field = cast(Self, cls()(attr))
        else:
            field = cls(default=attr)

        field.python_name = name
        # store origin django type for further usage
        if name in origin.__dict__.get("__annotations__", {}):
            field.origin_django_type = django_type

        # annotation of field is used as a class type
        if type_annotation is not None:
            field.type_annotation = type_annotation
            field.is_auto = is_auto(field.type_annotation)

        # resolve the django_name and check if it is relation field. django_name
        # is used to access the field data in resolvers
        try:
            model_field = get_model_field(django_type.model, field.django_name or name)
        except FieldDoesNotExist:
            model_attr = getattr(django_type.model, name, None)
            if model_attr is not None and isinstance(model_attr, ModelProperty):
                if field.is_auto:
                    field.type_annotation = StrawberryAnnotation(model_attr.type_annotation)
                    field.is_auto = is_auto(field.type_annotation)

                if field.description is None:
                    field.description = model_attr.description
            elif field.django_name or field.is_auto:
                raise  # field should exist, reraise caught exception
        else:
            field.is_relation = model_field.is_relation
            field.django_name = resolve_model_field_name(
                model_field,
                is_input=django_type.is_input,
                is_filter=django_type.is_filter,
            )

            # change relation field type to auto if field is inherited from another
            # type. for example if field is inherited from output type but we are
            # configuring field for input type
            if field.is_relation and not is_similar_django_type(
                django_type, field.origin_django_type
            ):
                field.is_auto = True

            # resolve type of auto field
            if field.is_auto:
                field_type = resolve_model_field_type(model_field, django_type)
                field.type_annotation = StrawberryAnnotation(field_type)

            if is_optional(model_field, django_type.is_input, django_type.is_partial):
                assert field.type_annotation
                field.type_annotation.annotation = Optional[
                    field.type_annotation.annotation  # type:ignore
                ]

            if field.description is None:
                description = (
                    model_field.field.help_text
                    if isinstance(model_field, (ManyToOneRel, ManyToManyRel))
                    else model_field.help_text
                )
                if description:
                    field.description = str(description)

        return field

    def get_result(
        self,
        source: Optional[models.Model],
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> Union[Awaitable[Any], Any]:
        if self.base_resolver is not None:
            # Unlike strawberry_django, we don't enforce this on sync_to_async since it adds
            # a lot of overhead which might be unnecessary. Leave it up to the implementation
            # to decide what to do...
            result = self.base_resolver(*args, **kwargs)
        elif source is None:
            result = self.model.objects.all()
        else:
            # Small optimization to async resolvers avoid having to call it in an sync_to_async
            # context if the value is already cached, since it will not hit the db anymore
            attname = self.django_name or self.python_name
            attr = getattr(source.__class__, attname, None)
            try:
                if isinstance(attr, DeferredAttribute):
                    result = source.__dict__[attr.field.attname]
                elif isinstance(attr, ModelProperty):
                    result = source.__dict__[attr.name]
                elif isinstance(attr, ForwardManyToOneDescriptor):
                    # This will raise KeyError if it is not cached
                    result = attr.field.get_cached_value(source)  # type:ignore
                elif isinstance(attr, ReverseOneToOneDescriptor):
                    # This will raise KeyError if it is not cached
                    result = attr.related.get_cached_value(source)
                elif isinstance(attr, ReverseManyToOneDescriptor):
                    # This returns a queryset, it is async safe
                    result = getattr(source, attname)
                else:
                    raise KeyError
            except KeyError:
                result = resolvers.getattr_async_unsafe(source, attname)

        if self.is_list:
            qs_resolver = lambda qs: self.get_queryset_as_list(qs, info, **kwargs)
        else:
            qs_resolver = lambda qs: self.get_queryset_one(qs, info, **kwargs)

        return resolvers.resolve_result(result, info=info, qs_resolver=qs_resolver)

    @resolvers.async_unsafe
    def get_queryset_as_list(self, qs: QuerySet[_M], info: Info, **kwargs) -> List[_M]:
        need_perm_safe = False
        for d in HasPermDirective.for_origin(self):
            qs = d.get_queryset(
                qs,
                info.context.request.user,
                ctype=ContentType.objects.get_for_model(qs.model),
            )
            need_perm_safe = need_perm_safe or is_perm_safe(qs)

        retval = list(self.get_queryset(qs, info, **kwargs))
        if need_perm_safe:
            retval = perm_safe(retval)

        return retval

    @resolvers.async_unsafe
    def get_queryset_one(self, qs: QuerySet[_M], info: Info, **kwargs) -> _M:
        return self.get_queryset(qs, info, **kwargs).get()


@overload
def field(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Any = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
) -> _T:
    ...


@overload
def field(
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Any = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
) -> Any:
    ...


@overload
def field(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Any = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
) -> StrawberryDjangoField:
    ...


def field(
    resolver=None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Any = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = UNSET,
    default_factory: Union[Callable, object] = UNSET,
    directives: Optional[Sequence[StrawberrySchemaDirective]] = (),
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init=None,
) -> Any:
    """Annotate a method or property as a Django GraphQL field.

    Examples:
        It can be used both as decorator and as a normal function:

        >>> @gql.type
        >>> class X:
        ...     field_abc: str = gql.django.field(description="ABC")
        ...     @gql.django.field(description="ABC")
        ...
        ...     def field_with_resolver(self) -> str:
        ...         return "abc"

    """
    f = StrawberryDjangoField(
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        directives=directives,
        filters=filters,
        only=only,
        select_related=select_related,
        prefetch_related=prefetch_related,
    )
    if resolver:
        f = f(resolver)
    return f



# Based on https://github.com/flavors/django-graphiql-debug-toolbar

import collections
import json
from typing import Optional
import weakref

from debug_toolbar.middleware import DebugToolbarMiddleware as _DebugToolbarMiddleware
from debug_toolbar.middleware import _HTML_TYPES
from debug_toolbar.middleware import show_toolbar
from debug_toolbar.panels.sql import panel, tracking
from debug_toolbar.toolbar import DebugToolbar
from django.core.serializers.json import DjangoJSONEncoder
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from strawberry.django.views import BaseView

_store_cache = weakref.WeakKeyDictionary()
_original_store = DebugToolbar.store


def _store(toolbar: DebugToolbar):
    _original_store(toolbar)
    _store_cache[toolbar.request] = toolbar.store_id


def _get_payload(request: HttpRequest, response: HttpResponse):
    store_id = _store_cache.get(request)
    if not store_id:
        return None

    toolbar: Optional[DebugToolbar] = DebugToolbar.fetch(store_id)
    if not toolbar:
        return None

    content = force_str(response.content, encoding=response.charset)
    payload = json.loads(content, object_pairs_hook=collections.OrderedDict)
    payload["debugToolbar"] = collections.OrderedDict([("panels", collections.OrderedDict())])
    payload["debugToolbar"]["storeId"] = toolbar.store_id

    for p in reversed(toolbar.enabled_panels):
        if p.panel_id == "TemplatesPanel":
            continue

        if p.has_content:
            title = p.title
        else:
            title = None

        sub = p.nav_subtitle
        payload["debugToolbar"]["panels"][p.panel_id] = {
            "title": title() if callable(title) else title,
            "subtitle": sub() if callable(sub) else sub,
        }

    return payload


DebugToolbar.store = _store  # type:ignore


def _wrap_cursor(connection, panel):
    c = type(connection)
    if hasattr(c, "_djdt_cursor"):
        return None

    c._djdt_cursor = c.cursor
    c._djdt_chunked_cursor = c.chunked_cursor

    def cursor(*args, **kwargs):
        return tracking.state.Wrapper(c._djdt_cursor(*args, **kwargs), args[0], panel)

    def chunked_cursor(*args, **kwargs):
        cursor = c._djdt_chunked_cursor(*args, **kwargs)
        if not isinstance(cursor, tracking.BaseCursorWrapper):
            return tracking.state.Wrapper(cursor, args[0], panel)
        return cursor

    c.cursor = cursor
    c.chunked_cursor = chunked_cursor

    return cursor


def _unwrap_cursor(connection):
    c = type(connection)
    if not hasattr(c, "_djdt_cursor"):
        return

    c.cursor = c._djdt_cursor
    c.chunked_cursor = c._djdt_chunked_cursor
    del c._djdt_cursor
    del c._djdt_chunked_cursor


# Patch wrap_cursor/unwrap_cursor so that they work with async views
# Are there any drawbacks to this?
tracking.wrap_cursor = _wrap_cursor
tracking.unwrap_cursor = _unwrap_cursor
panel.wrap_cursor = _wrap_cursor
panel.unwrap_cursor = _unwrap_cursor


class DebugToolbarMiddleware(_DebugToolbarMiddleware):
    sync_capable = True
    async_capable = True

    def __call__(self, request: HttpRequest):
        response = super().__call__(request)

        if not show_toolbar(request) or DebugToolbar.is_toolbar_request(request):
            return response

        content_type = response.get("Content-Type", "").split(";")[0]
        is_html = content_type in _HTML_TYPES
        is_graphiql = getattr(request, "_is_graphiql", False)

        if is_html and is_graphiql and response.status_code == 200:
            template = render_to_string("strawberry_django_plus/debug_toolbar.html")
            response.write(template)
            if "Content-Length" in response:
                response["Content-Length"] = len(response.content)

        if is_html or not is_graphiql or content_type != "application/json":
            return response

        payload = _get_payload(request, response)
        if payload is None:
            return response

        response.content = json.dumps(payload, cls=DjangoJSONEncoder)
        if "Content-Length" in response:
            response["Content-Length"] = len(response.content)

        return response

    def process_view(self, request: HttpRequest, view_func, *args, **kwargs):
        view = getattr(view_func, "view_class", None)
        request._is_graphiql = bool(view and issubclass(view, BaseView))  # type:ignore



from strawberry_django import (
    DjangoFileType,
    DjangoImageType,
    DjangoModelType,
    ManyToManyInput,
    ManyToOneInput,
    OneToManyInput,
    OneToOneInput,
    auth,
    filter,
    filters,
    mutation,
    mutations,
    ordering,
    types,
)

from strawberry_django_plus.descriptors import model_cached_property, model_property
from strawberry_django_plus.field import field
from strawberry_django_plus.mutations import input_mutation
from strawberry_django_plus.type import input, interface, type
from strawberry_django_plus.utils.resolvers import async_unsafe, resolve_qs

__all__ = [
    # strawberry_django
    "auth",
    "filters",
    "filter",
    "ordering",
    "types",
    "DjangoFileType",
    "DjangoImageType",
    "DjangoModelType",
    "OneToOneInput",
    "OneToManyInput",
    "ManyToOneInput",
    "ManyToManyInput",
    "mutations",
    "mutation",
    # strawberry_django_plus
    "field",
    "async_unsafe",
    "resolve_qs",
    "type",
    "interface",
    "input",
    "model_property",
    "model_cached_property",
    "input_mutation",
]
